import torch
from .ActivationFunction import ActivationFunction


class Gish(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        a = -torch.exp(x)
        return x * torch.log(2 - torch.exp(a))



import torch


class ParametricLogish(torch.nn.Module):
    def __init__(self, alpha: float = 1.0, beta: float = 10.0):
        super().__init__()
        self.alpha = alpha
        self.beta = beta

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = torch.sigmoid(self.beta * x)
        b = torch.log(1 + a)
        return self.alpha * x * b



from ..activations import (
    AOAF,
    AReLU,
    DPReLU,
    DualLine,
    FReLU,
    LeLeLU,
    PERU,
    PiLU,
    ShiLU,
    StarReLU,
)


class StochasticClassifier:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._activation_pool = [
            AOAF,
            AReLU,
            DPReLU,
            DualLine,
            FReLU,
            LeLeLU,
            PERU,
            PiLU,
            ShiLU,
            StarReLU,
        ]



import torch
from torchvision.models import resnet50
from torch.nn.init import kaiming_uniform_ as kaiming


class ResnetModule(torch.nn.Module):
    def __init__(self, num_classes: int):
        super().__init__()
        self.num_classes = num_classes
        self._create_encoder()
        self._create_fc()

    def _create_encoder(self):
        self.encoder = resnet50(weights="DEFAULT")
        self.encoder.fc = torch.nn.Identity()

    def _create_fc(self):
        self.fc = torch.nn.Linear(2048, self.num_classes)
        kaiming(self.fc.weight, mode="fan_in", nonlinearity="relu")

    def forward(self, images: torch.Tensor):
        features = self.encoder(images)
        return self.fc(features)



from .concerns import (
    ReportsSize,
    StoresModels,
    ReportsClassificationMetrics,
    PredictsProbabilities,
)


class Ensemble(
    ReportsSize,
    ReportsClassificationMetrics,
    PredictsProbabilities,
    StoresModels,
):
    def __init__(self):
        super().__init__()



import torch
import math
from .ActivationFunction import ActivationFunction


class SGELU(ActivationFunction):
    def __init__(self, alpha: float = 0.1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha = alpha

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.alpha * x * torch.erf(x / math.sqrt(2))



import torch
from .ActivationFunction import ActivationFunction


class SoftsignRReLU(ActivationFunction):
    def __init__(self, lower: float = 1 / 8, upper: float = 1 / 3, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lower = lower
        self.upper = upper

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = torch.empty_like(x).uniform_(self.lower, self.upper)
        return torch.where(
            x >= 0,
            torch.div(1, (1 + x).pow(2)) + x,
            torch.div(1, (1 + x).pow(2)) + a * x,
        )



import torch
from .ActivationFunction import ActivationFunction


class AdaptiveActivationFunction(ActivationFunction):
    def __init__(self):
        super().__init__()

    def __repr__(self):
        arguments_text = ""

        if hasattr(self, "num_channels"):
            arguments_text = f"channels={self.num_channels}"

        return f"{__class__.__name__}({arguments_text})"



import torch
from .ActivationFunction import ActivationFunction


class Logish(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.log1p(torch.sigmoid(x))
